/******************************************************************************
*
* vdoxs.c
*
* Virtual Driver Example - OXS
*
*  Copyright 2011-2014 Citrix Systems, Inc. All Rights Reserved.
*
* Citrix hereby permits you to use, reproduce, modify, display, compile,
* sublicense, and distribute all or any portion of the sample code contained
* in this file, provided that you remove this Copyright attribution from any
* modified version of this file.
*
* THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
*
*******************************************************************************
*
*   OXS Description:
*
*   The OXS (a.k.a. Noughts & Crosses) virtual channel implements remote
*   drawing of a Noughts & Crosses game. The purpose of this is to
*   demonstrate three Linux Receiver specific interfaces:
*       * Sub-window/overlay buffer
*       * Events (File Descriptor selecting)
*       * Timers
*
*   The client side component the OXS virtual driver implments the drawing of
*   the Noughts & Crosses game. All processing of the game play itself is
*   handled by the server side. To achieve client side drawing a
*   sub-window (child) of the session window is obtained using the sub-window
*   interface. The obtained sub-window is placed over the corresponding area of
*   the Noughts & Crosses game window and every movement to the server side
*   game window is mimicked by the sub-window. This creates the impression of a
*   single application. With clicks and interaction with the server side
*   OXS application converted into Play Move, Winning Line, Move Window, and
*   Close Window commands. These commands are received by the virtual driver
*   and translated into the corresponding actions and X drawing commands.
*
*   This sub-window technology is most useful for solutions such as local video
*   decoding. The sub-window interface at present is not designed to take
*   keyboard and mouse input, and is simply for rendering graphics.
*
*   The OXS virtual driver also uses the events (Evt) interface. This is used
*   to monitor the X server file descriptor for expose events. Allowing a
*   callback to re-draw the game area on every expose event.
*
*   The OXS virtual driver uses the timer (Tmr) interface as well. This is used
*   to work around a race condition between the OXS and Seamless virtual
*   channels. As mentioned previously the sub-window used for drawing mimicks
*   the movements of the server side window. However, if Seamless is in use
*   there is a race condition between the movement of the Seamless server side
*   window and the sub-window. The timer is used to delay the sub-window
*   position update in order to ensure the Seamless window move has been
*   completed first. This is a workaround and needs to be fixed.
*
******************************************************************************/
#define X11_SUPPORT     /* Remove for alternative video output method. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "citrix.h"
#include "clib.h"

/* from src/shared/inc/citrix */
#include "ica.h"
#include "ica-c2h.h"
#include "vdoxs.h"

/* from src/inc */
#include "icaconst.h"
#include "clterr.h"
#include "wdapi.h"
#include "vdapi.h"
#include "vd.h"
#include "ctxdebug.h"
#include "logapi.h"
#include "miapi.h"
#include "ctxnevt.h"
#include "ctxntmr.h"

#include "cwire.h"
#include "cmacro.h"
#include "mainloop.h"

#ifdef X11_SUPPORT
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#endif /* X11_SUPPORT */

#include "subwindow.h"

#define INI_OXS_VDSECTION "OXS"

/* make sure that we pull in Load from vdapi.a */
extern int Load();

struct _wrapload {
        int ((*fn)());
} WRAP_LOAD[] = {
        {Load}
};


/*=============================================================================
==   Functions Defined
=============================================================================*/
/*
 *   These are the required driver functions (exported)
 */

int DriverOpen( PVD, PVDOPEN, PUINT16 );
int DriverClose( PVD, PDLLCLOSE, PUINT16 );
int DriverInfo( PVD, PDLLINFO, PUINT16 );
int DriverPoll( PVD, PDLLPOLL, PUINT16 );
int DriverQueryInformation( PVD, PVDQUERYINFORMATION, PUINT16 );
int DriverSetInformation( PVD, PVDSETINFORMATION, PUINT16 );
int DriverGetLastError( PVD, PVDLASTERROR );

static int WFCAPI ICADataArrival( PVOID, USHORT, LPBYTE, USHORT );


/*=============================================================================
==   Data
=============================================================================*/

/*
 *  These are returned when we register our hook
 */

PVOID               pWd             = NULL;
POUTBUFRESERVEPROC  pOutBufReserve  = NULL;
POUTBUFAPPENDPROC   pOutBufAppend   = NULL;
POUTBUFWRITEPROC    pOutBufWrite    = NULL;
PAPPENDVDHEADERPROC pAppendVdHeader = NULL;


/*=============================================================================
==   Static Data
=============================================================================*/
#ifdef X11_SUPPORT
static Display     *disp;                 /* X display connection */
static GC           gc;                   /* Graphics Context for line drawing */
static PEVT         fdEvt;                /* File descriptor select callback structure for X */
static BOOLEAN      parentReady = FALSE;  /* Is the parent window ready to be used */
static PTMR         windowMoveTimer;      /* Timer callback structure for delaying window move */

#else

static UINT32      *overlay_buffer;       /* Assumes 4 bytes/pixel. */
static int          buffer_stride;        /* Pixels. */
#endif /* X11_SUPPORT */

static UINT32       xwin = 0;             /* X window ID of client sub window */
static SIGNED_RECT  windowGeom;           /* Store window position information */
static UINT32       parent = 0;           /* Server window (also seamless window) parent id */

static USHORT       gusMaxDataSize;       /* Maximum Data Write Size */
static USHORT       gusVirtualChannelNum; /* Channel number assigned by WD */


/* Game state */
static enum counter gridState[OXS_GRID_WIDTH][OXS_GRID_HEIGHT] = {
    {NONE, NONE, NONE}, {NONE, NONE, NONE}, {NONE, NONE, NONE}
};
static BOOLEAN      winLineVisible = FALSE;
static OXS_WINNING_LINE_PACKET winLine;


/*=============================================================================
==   Local Functions
=============================================================================*/

#ifdef X11_SUPPORT
/*-----------------------------------------------------------------------------
--   X drawing and event handling functions
-----------------------------------------------------------------------------*/

static void initX(void)
{
    disp = XOpenDisplay(NULL);

    TRACE((TC_VD, TT_API3, "Initialise X: X window ID=%d, Display FD=%d", xwin, disp));

    /* Wipe the window background */
    XSetWindowBackground(disp, xwin, WhitePixel(disp, DefaultScreen(disp)));
    XClearArea(disp, xwin, 0, 0, 0, 0, True);

    {
        /* Create GC for drawing lines */
        XGCValues val;

        val.line_width = 3;
        gc = XCreateGC(disp, xwin, GCLineWidth, &val);
    }

    /* Monitor for expose events */
    XSelectInput(disp, xwin, ExposureMask);
}


/* Draw a nought counter */

static void drawNought(int x, int y)
{
    /* Draw a complete circle */
    XDrawArc(disp, xwin, gc, x, y, OXS_HALF_SQUARE, OXS_HALF_SQUARE, 0, 360*64);
}

/* Draw a line. */

static void draw_line(int s_x, int s_y, int d_x, int d_y)
{
    XDrawLine(disp, xwin, gc, s_x, s_y, d_x, d_y);
}

/* Push a "frame" to the screen. */

static void push_frame(void)
{
    XFlush(disp);
}
#else /* Not using X11. */

/* On reconnect, the application may be running, and can send play commands
 * without ever setting the geometry, which allocates the buffer.
 * So the drawing operations must check the buffer first.
 */

/* Set the visible part of the buffer to white. */

static void clear_buffer(void)
{
    int         i;
    char       *dp;

    dp = (char *)overlay_buffer;
    if (!dp)
        return;
    for (i = 0; i < OXS_GRID_HEIGHT * OXS_SQUARE_SIZE; ++i) {
        memset(dp, 0xff, OXS_GRID_WIDTH * OXS_SQUARE_SIZE * sizeof (UINT32));
        dp += buffer_stride * sizeof (UINT32);
    }
}

/* Push a change to the screen. */

static void push_frame(void)
{
    MM_push_overlay(xwin);
}

/* Draw a line. Only horizontal, vertical and 45-degree lines are used. */

static void draw_line(int s_x, int s_y, int d_x, int d_y)
{
    int         i, n, x_inc, y_inc;
    UINT32     *dp;

    if (!overlay_buffer)
        return;
    if (s_x < d_x) {
        x_inc = 1;
        n = d_x - s_x;
    } else if (s_x == d_x) {
        x_inc = 0;
        n = 0;
    } else {
        x_inc = -1;
        n = s_x - d_x;
    }

    if (s_y < d_y) {
        y_inc = 1;
        if (n == 0)
            n = d_y - s_y;
    } else if (s_y == d_y) {
        y_inc = 0;
    } else {
        y_inc = -1;
        if (n == 0)
        n = s_y - d_y;
    }

    dp = overlay_buffer + (s_y * buffer_stride) + s_x;
    for (i = 0; i < n; ++i) {
        *dp = 0;       /* Black. */
        dp += (y_inc * buffer_stride) + x_inc;
    }
}
    
/* Draw a nought counter crudely, a tilted square. */

static void drawNought(int x, int y)
{
    x += OXS_QUARTER_SQUARE;    /* Centre. */
    y += OXS_QUARTER_SQUARE;
    draw_line(x, y - OXS_QUARTER_SQUARE, x + OXS_QUARTER_SQUARE, y);
    draw_line(x + OXS_QUARTER_SQUARE, y, x, y + OXS_QUARTER_SQUARE);
    draw_line(x, y + OXS_QUARTER_SQUARE, x - OXS_QUARTER_SQUARE, y);
    draw_line(x - OXS_QUARTER_SQUARE, y, x, y - OXS_QUARTER_SQUARE);
}

#endif /* not X11_SUPPORT */

/* Common counter drawing commands. */

/* Draw a cross counter */

static void drawCross(int x, int y)
{
    draw_line(x, y, x + OXS_HALF_SQUARE, y + OXS_HALF_SQUARE);
    draw_line(x, y + OXS_HALF_SQUARE, x + OXS_HALF_SQUARE, y);
}

static void drawCounter(POXS_COORDINATE position, enum counter counter)
{
    int x, y;

    x = (position->x * OXS_SQUARE_SIZE) + OXS_QUARTER_SQUARE;
    y = (position->y * OXS_SQUARE_SIZE) + OXS_QUARTER_SQUARE;

    switch (counter)
    {
        case NOUGHT:
            drawNought(x, y);
            break;
        case CROSS:
            drawCross(x, y);
            break;
        case NONE:
            break;
    }
}


/* Draw winning line from the centre of two grid squares */
static void drawWinningLine(POXS_COORDINATE start, POXS_COORDINATE finish)
{
    draw_line(start->x*OXS_SQUARE_SIZE + OXS_HALF_SQUARE,
              start->y*OXS_SQUARE_SIZE + OXS_HALF_SQUARE,
              finish->x*OXS_SQUARE_SIZE + OXS_HALF_SQUARE,
              finish->y*OXS_SQUARE_SIZE + OXS_HALF_SQUARE);
}


/* Draw the game grid, all placed counters and winning line */
static void drawGame(void)
{
    int i, j;

    /* Draw vertical grid lines */
    for (i = 1; i < OXS_GRID_WIDTH; i++)
        draw_line(i*OXS_SQUARE_SIZE, 0,
                  i*OXS_SQUARE_SIZE, OXS_GRID_HEIGHT*OXS_SQUARE_SIZE);

    /* Draw horizontal grid lines */
    for (j = 1; j < OXS_GRID_HEIGHT; j++)
        draw_line(0, j*OXS_SQUARE_SIZE,
                  OXS_GRID_WIDTH*OXS_SQUARE_SIZE, j*OXS_SQUARE_SIZE);

    /* Draw all placed counters */
    for (i = 0; i<OXS_GRID_WIDTH; i++)
    {
        for (j = 0; j<OXS_GRID_HEIGHT; j++)
        {
            if (gridState[i][j] != NONE)
            {
                OXS_COORDINATE p;
                p.x = i;
                p.y = j;
                drawCounter(&p, gridState[i][j]);
            }
        }
    }

    /* Draw winning line */
    if(winLineVisible)
        drawWinningLine(&winLine.startpos, &winLine.finishpos);
}


/* Wipe any game state */
static void cleanGame(void)
{
    int i, j;
    for (i = 0; i<OXS_GRID_WIDTH; i++)
    {
        for (j = 0; j<OXS_GRID_HEIGHT; j++)
        {
            gridState[i][j] = NONE;
        }
    }

    winLineVisible = FALSE;
}


/*-----------------------------------------------------------------------------
--   Linux Receiver Timer, Event, Seamless window creation callbacks
-----------------------------------------------------------------------------*/

#ifdef X11_SUPPORT
/* X event file descriptor callback. This is setup using the Linux Receivers
   event (Evt) interface. In this case it is monitoring the X file descriptor */
static void call_Xevt(PVOID pSubscriberId, PVOID pHnd)
{
    XEvent e;

    /* Simple X event loop */
    while(XPending(disp))
    {
        XNextEvent(disp, &e);

        switch (e.type)
        {

            /* Detect Expose events */
            case Expose:
                TRACE((TC_VD, TT_API3, "Expose event detected re-drawing game"));
                drawGame();
        }
    }
}

/* Window move delay callback function. This is setup using the Linux Receivers
   timer (Tmr) interface. */
static void windowMoveDelay(PVOID pvSubscriber, PTMR phTimer)
{
    PSIGNED_RECT w = (PSIGNED_RECT) pvSubscriber;

    TRACE((TC_VD, TT_API3, "Window Move delay finished, moving window: (%d,%d) (%d,%d)",
           w->left, w->top, w->right, w->bottom));

    /* Set the new subwindow position */
    MM_set_geometry(parent, xwin, w);

    /* Disable the timer as it is recurrent by default */
    Tmr_setEnabled(windowMoveTimer, FALSE);
}


/* Seamless window creation callback. This is part of the Linux Receiver's
 * subwindow interface and is used to avoid race conditions between
 * the creation of seamless windows and their use.
 */

static void seamlessWindowCreated(UINT32 hwin)
{
    if (parent == hwin)
    {
        /* The seamless window we require has been created! */
        TRACE((TC_VD, TT_API3, "Required seamless window created!"));

        /* Start the windowMoveTimer */
        Tmr_setEnabled(windowMoveTimer, TRUE);

        parentReady = TRUE;
     
        /* Remove the seamless window creation callback */
        MM_TWI_clear_new_window_function(seamlessWindowCreated);
    }
}
#endif /* X11_SUPPORT */

/*-----------------------------------------------------------------------------
--   OXS packet processing functions
-----------------------------------------------------------------------------*/

/* Set the server parent window ID */
static void processParentWindowPacket(LPBYTE pBuf)
{
    POXS_PARENT_WINDOW_PACKET p;
    p = (POXS_PARENT_WINDOW_PACKET) (pBuf + sizeof(OXS_HEADER));

    TRACE((TC_VD, TT_API3, "Parent Window: 0x%x", p->parent));

    /* Set the servers parent window for the subwindow.
       This is useful when the windows are seamless. */
    parent = p->parent;
}


/* Draw a counter to screen */
static void processPlayMovePacket(LPBYTE pBuf)
{
    POXS_PLAY_MOVE_PACKET p;
    p = (POXS_PLAY_MOVE_PACKET) (pBuf + sizeof(OXS_HEADER));

    TRACE((TC_VD, TT_API3, "Play Move: position=(%d,%d) counter=%d\n",
           p->position.x, p->position.y, p->counter));

    gridState[p->position.x][p->position.y] = p->counter;
    drawGame();
    push_frame();
}


/* Draw the winning line between two grid squares */
static void processWinningLinePacket(LPBYTE pBuf)
{
    POXS_WINNING_LINE_PACKET p;

    p = (POXS_WINNING_LINE_PACKET) (pBuf + sizeof(OXS_HEADER)); 

    TRACE((TC_VD, TT_API3, "Winning Line: start position=(%d,%d) finish position=(%d,%d)",
           p->startpos.x, p->startpos.y, p->finishpos.x, p->finishpos.y));

    drawWinningLine(&p->startpos, &p->finishpos);
    push_frame();

    /* Store winning line for subsequent re-draws */
    memcpy(&winLine, p, sizeof(OXS_WINNING_LINE_PACKET));

    winLineVisible = TRUE;
}


/* Move client side subwindow as instructed by the server */
static void processMoveWindowPacket(LPBYTE pBuf)
{  
    POXS_MOVE_WINDOW_PACKET pMonitorInfo;
    pMonitorInfo = (POXS_MOVE_WINDOW_PACKET) (pBuf + sizeof(OXS_HEADER));

    windowGeom.left = pMonitorInfo->x;
    windowGeom.top = pMonitorInfo->y;
    windowGeom.right = pMonitorInfo->x + OXS_GRID_WIDTH*OXS_SQUARE_SIZE;
    windowGeom.bottom = pMonitorInfo->y + OXS_GRID_HEIGHT*OXS_SQUARE_SIZE;

    TRACE((TC_VD, TT_API3, "Move Window: store position (%d,%d) (%d,%d)",
           windowGeom.left, windowGeom.top, windowGeom.right, windowGeom.bottom));

#ifdef X11_SUPPORT
    if (parent != 0)
    {
        /* We are guaranteed to get a OXS_PARENT_WINDOW before
           the last OXS_MOVE_WINDOW packet */

        if (!xwin)
        {
            /* Request client window & open X display connection */

            if (!MM_get_window(parent, &xwin))
            {
                TRACE((TC_VD, TT_API3, "Seamless window 0x%x does not yet exist!", parent));
                /* The seamless window xwin is not ready yet. Setup a callback to get notified
                   when it is ready */
                MM_TWI_set_new_window_function(seamlessWindowCreated);
            }
            else
            {
                TRACE((TC_VD, TT_API3, "Seamless window 0x%x parent ready", parent));
                parentReady = TRUE;
            }

            /* Initialise graphics output. */

            initX();

            /* Setup File Descriptor select on the X server connection */
            Evt_create(NULL, (PFNDELIVER) call_Xevt, NULL, &fdEvt);
            Evt_trigger_for_input(fdEvt, ConnectionNumber(disp));

            /* Draw the grid for the first time */

            drawGame();

            /* Push out the drawn frame. */

            push_frame();
        }

        /* Set the client side window position and size. If the parent is ready. */
        if (parentReady)
        {
            /* This timer is only required to work around a race condition
             * between the seamless virtual channel and the OXS channel
             * regarding window positioning. If the OXS channel moves
             * the subwindow before the seamless window has moved then
             * the subwindow will be out of place.
             */

            /* Start/Restart the windowMoveTimer */
            Tmr_setEnabled(windowMoveTimer, FALSE);
            Tmr_setEnabled(windowMoveTimer, TRUE);
        }
    }
#else /* Overlay buffer, not X11. */
    /* Overlay buffers are less sensitive to timing.
     * Set the geometry and draw directly.
     */

    if (!xwin)
        MM_get_overlay_buffer(parent, &xwin, MM_BUF_STRIDE_OK, NULL);
    MM_set_geometry(parent, xwin, &windowGeom);
    {
        const struct special_buffer *bip;

        bip = MM_get_buffer(xwin);
        if (bip) {
            overlay_buffer = (UINT32 *)bip->buf_addr;
            buffer_stride = bip->params.overlay_buffer.buffer.width;
        } else {
            overlay_buffer = NULL;
        }
    }
    clear_buffer();
    drawGame();
    push_frame();
#endif /* not X11_SUPPORT */            
}


/* Close the client side subwindow and tidy up */

static void processWindowClosePacket(LPBYTE pBuf)
{
    TRACE((TC_VD, TT_API3, "Close Window"));
    if (xwin)
    {
        /* Destroy the client side subwindow */
        MM_destroy_window(parent, xwin);
        xwin = 0;

        /* Clean up any game state */
        cleanGame();
    }
    parent = 0;
}


/*******************************************************************************

 *
 *  ICADataArrival
 *
 *   A data PDU arrived over our channel.
 *

 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *
 *    uChan (input)
 *       ICA channel the data is for.

 *
 *    pBuf (input)
 *       Buffer with arriving data packet
 *
 *    Length (input)
 *       Length of the data packet
 *

 * EXIT:
 *       void
 *
 ******************************************************************************/
static int WFCAPI ICADataArrival(PVOID pVd, USHORT uChan, LPBYTE pBuf,
                                  USHORT Length)
{
    POXS_HEADER h = (POXS_HEADER) pBuf;

    TRACE((TC_VD, TT_API3, "VDOXS: IcaDataArrival, Len=%d, type=%d", Length, h->type));
    TRACEBUF((TC_VD, TT_API3, pBuf, Length));

    switch (h->type)
    {
        case OXS_PARENT_WINDOW:
            processParentWindowPacket(pBuf);
            break;
        case OXS_PLAY_MOVE:
            processPlayMovePacket(pBuf);
            break;
        case OXS_WINNING_LINE:
            processWinningLinePacket(pBuf);
            break;
        case OXS_MOVE_WINDOW:
            processMoveWindowPacket(pBuf);
            break;
        case OXS_CLOSE_WINDOW:
            processWindowClosePacket(pBuf);
            break;
        default:
            TRACE((TC_VD, TT_API3, "Unrecognised Packet!"));
            break;           
    }

    return CLIENT_STATUS_SUCCESS;
}


/*=============================================================================
==   Global Functions
=============================================================================*/

/*******************************************************************************
 *
 *  DriverOpen
 *
 *    Called once to set up things.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdOpen (input/output)
 *       pointer to the structure VDOPEN
 *    puiSize (Output)
 *       size of VDOPEN structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *    CLIENT_ERROR_NO_MEMORY - could not allocate data buffer
 *    On other errors, an error code is returned from lower level functions.
 *
 ******************************************************************************/
int DriverOpen(PVD pVd, PVDOPEN pVdOpen, PUINT16 puiSize)
{
    WDSETINFORMATION   wdsi;
    VDWRITEHOOK        vdwh;
    WDQUERYINFORMATION wdqi;
    OPENVIRTUALCHANNEL OpenVirtualChannel;
    int                rc;
    UINT16             uiSize;

    TRACE((TC_VD, TT_API1, "VDOXS: DriverOpen entered"));

    *puiSize = sizeof(VDOPEN);

    /*
     * Get a virtual channel
     */
    wdqi.WdInformationClass = WdOpenVirtualChannel;
    wdqi.pWdInformation = &OpenVirtualChannel;
    wdqi.WdInformationLength = sizeof(OPENVIRTUALCHANNEL);
    OpenVirtualChannel.pVCName = CTXOXS_VIRTUAL_CHANNEL_NAME;

    /* uiSize will be set when we return back from VdCallWd. */
    uiSize = sizeof(WDQUERYINFORMATION);

    rc = VdCallWd(pVd, WDxQUERYINFORMATION, &wdqi, &uiSize);
    TRACE((TC_VD, TT_API1, "VDOXS: opening channel %s",
           OpenVirtualChannel.pVCName));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE((TC_VD, TT_ERROR, "VDOXS: Could not open %s. rc=%d.",
               OpenVirtualChannel.pVCName, rc));
        return rc;
    }

    gusVirtualChannelNum = OpenVirtualChannel.Channel;

    /*
     *  Used by the engine
     */
    pVdOpen->ChannelMask = (1L << gusVirtualChannelNum);
    pVd->pPrivate = NULL; /* pointer to private data, if needed */

    /*
     *  Register write hooks for our virtual channel
     */
    vdwh.Type = gusVirtualChannelNum;
    vdwh.pVdData = pVd;
    vdwh.pProc = ICADataArrival;
    wdsi.WdInformationClass  = WdVirtualWriteHook;
    wdsi.pWdInformation      = &vdwh;
    wdsi.WdInformationLength = sizeof(VDWRITEHOOK);
    uiSize                   = sizeof(WDSETINFORMATION);

    rc = VdCallWd(pVd, WDxSETINFORMATION, &wdsi, &uiSize);
    TRACE((TC_VD, TT_API2, "VDOXS: writehook channel number=%u"
           " vdwh.pVdData=%x rc=%d", vdwh.Type, vdwh.pVdData, rc));

    if (rc != CLIENT_STATUS_SUCCESS) {
        TRACE((TC_VD, TT_ERROR, "VDOXS: Could not register write hook. rc %d", rc));
        return rc;
    }

    /*
     *  Grab pointers to functions to use to send data to the host
     */
    pWd            = vdwh.pWdData;
    pOutBufReserve = vdwh.pOutBufReserveProc;
    pOutBufAppend  = vdwh.pOutBufAppendProc;
    pOutBufWrite   = vdwh.pOutBufWriteProc;
    pAppendVdHeader= vdwh.pAppendVdHeaderProc;

    TRACE((TC_VD, TT_API1, "VDOXS: Registered"));

#ifdef X11_SUPPORT
    /* Allocate and initialise timer structure windowGeom. Which will be used
       to avoid race condition between seamless window moves and subwindow
       moves. This is a workaround and needs to be fixed. */
    Tmr_create(NULL, 500, (PVOID) &windowGeom,
               (void (*)(void *, void *)) windowMoveDelay, &windowMoveTimer);
#endif
    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverPoll
 *
 *  The Winstation driver calls DriverPoll
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdPoll (input)
 *       pointer to the structure DLLPOLL  (not used)
 *    puiSize (input)
 *       size of DLLPOOL structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - OK so far.  We will be polled again.
 *    CLIENT_STATUS_NO_DATA - No more data to send.
 *    CLIENT_STATUS_ERROR_RETRY - Could not get an output buffer.
 *                            Hopefully, one will be available next time.
 *    Otherwise, a fatal error code is returned from lower level functions.
 *
 *    NOTE:  CLIENT_STATUS_NO_DATA signals the WD that it is OK to stop
 *           polling until another host to client packet comes through.
 ******************************************************************************/
int DriverPoll(PVD pVd, PDLLPOLL pVdPoll, PUINT16 puiSize)
{
    /* This example doesn't have any client to server packets */
    return CLIENT_STATUS_NO_DATA;
}


/*******************************************************************************
 *
 *  DriverClose
 *
 *  The user interface calls VdClose to close a Vd before unloading it.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to procotol driver data structure
 *    pVdClose (input/output)
 *       pointer to the structure DLLCLOSE
 *    puiSize (input)
 *       size of DLLCLOSE structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int DriverClose(PVD pVd, PDLLCLOSE pVdClose, PUINT16 puiSize)
{
    TRACE(( TC_VD, TT_API1, "VDOXS: DriverClose entered" ));

    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverInfo
 *
 *    This routine is called to get module information
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdInfo (output)
 *       pointer to the structure DLLINFO
 *    puiSize (output)
 *       size of DLLINFO structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int DriverInfo(PVD pVd, PDLLINFO pVdInfo, PUINT16 puiSize)
{
    USHORT ByteCount;
    PVDOXS_C2H pVdData;
    PMODULE_C2H pHeader;
    PVDFLOW pFlow;

    ByteCount = sizeof(VDOXS_C2H);

    *puiSize = sizeof(DLLINFO);

    TRACE((TC_VD, TT_API1, "VDOXS: DriverInfo entered"));

    /*
     *  Check if buffer is big enough
     *  If not, the caller is probably trying to determine the required
     *  buffer size, so return it in ByteCount.
     */
    if (pVdInfo->ByteCount < ByteCount) {
        pVdInfo->ByteCount = ByteCount;
        return CLIENT_ERROR_BUFFER_TOO_SMALL;
    }

    /*
     *  Initialize default data
     */
    pVdInfo->ByteCount = ByteCount;
    pVdData = (PVDOXS_C2H) pVdInfo->pBuffer;

    /*
     *  Initialize module header
     */
    pHeader = &pVdData->Header.Header;
    pHeader->ByteCount = ByteCount;
    pHeader->ModuleClass = Module_VirtualDriver;

    pHeader->VersionL = CTXOXS_VER_LO;
    pHeader->VersionH = CTXOXS_VER_HI;
    strcpy((char*)(pHeader->HostModuleName), "ICA"); /* max 8 characters */

    /*
     *  Initialize virtual driver header
     */
    pVdData->Header.ChannelMask = pVd->ChannelMask;
    pFlow = &pVdData->Header.Flow;
    pFlow->BandwidthQuota = 0;
    pFlow->Flow = VirtualFlow_None;

    /* add our own data */
    pVdData->usMaxDataSize = gusMaxDataSize;

    pVdInfo->ByteCount = WIRE_WRITE(VDOXS_C2H, pVdData, ByteCount);

    return CLIENT_STATUS_SUCCESS;
}



/*******************************************************************************
 *
 *  DriverQueryInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdQueryInformation (input/output)
 *       pointer to the structure VDQUERYINFORMATION
 *    puiSize (output)
 *       size of VDQUERYINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int DriverQueryInformation(PVD pVd, PVDQUERYINFORMATION pVdQueryInformation, PUINT16 puiSize)
{
    TRACE(( TC_VD, TT_API1, "VDOXS: DriverQueryInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdQueryInformation->VdInformationClass = %d",
         pVdQueryInformation->VdInformationClass));

    *puiSize = sizeof(VDQUERYINFORMATION);

    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverSetInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdSetInformation (input/output)
 *       pointer to the structure VDSETINFORMATION
 *    puiSize (input)
 *       size of VDSETINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int DriverSetInformation(PVD pVd, PVDSETINFORMATION pVdSetInformation, PUINT16 puiSize)
{
    TRACE((TC_VD, TT_API1, "VDOXS: DriverSetInformation entered"));
    TRACE((TC_VD, TT_API2, "pVdSetInformation->VdInformationClass = %d",
           pVdSetInformation->VdInformationClass));

   return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverGetLastError
 *
 *   Queries error data.
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pLastError (output)
 *       pointer to the error structure to return (message is currently always
 *       NULL)
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int DriverGetLastError(PVD pVd, PVDLASTERROR pLastError)
{
    TRACE((TC_VD, TT_API1, "VDOXS: DriverGetLastError entered"));
    /*
     *  Interpret last error and pass back code and string data
     */

    pLastError->Error = pVd->LastError;
    pLastError->Message[0] = '\0';
    return CLIENT_STATUS_SUCCESS;
}
