/*************************************************************************
*
* VDPING.C
*
* Virtual Driver Example - Ping
*
* Copyright 1998-2014 Citrix Systems, Inc. All Rights Reserved.
*
* Citrix hereby permits you to use, reproduce, modify, display, compile,
* sublicense, and distribute all or any portion of the sample code contained
* in this file, provided that you remove this Copyright attribution from any
* modified version of this file.
*
* THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
*
*************************************************************************/

#ifdef DOS32
#define CITRIX 1
#include <sys/types.h>
#include "wfc32.h"
#endif

#ifndef unix
#include <windows.h>         /* for far etc. */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "citrix.h"
#include "clib.h"

/* from src/shared/inc/citrix */
#include "ica.h"
#include "ica-c2h.h"
#include "vdping.h"

/* from src/inc */
#include "icaconst.h"
#include "clterr.h"
#include "wdapi.h"
#include "vdapi.h"
#include "vd.h"
#include "ctxdebug.h"
#include "logapi.h"
#include "miapi.h"

#include "cwire.h"
#include "cmacro.h"

#if defined(macintosh) || defined(unix)
#include "pingwire.h"
#endif

#define INI_PING_VDSECTION "Ping"
#define INI_PING_PINGCOUNT "PingCount"
#define DEF_PING_PINGCOUNT 3

#ifdef unix
// make sure that we pull in Load from vdapi.a
extern int Load();

struct _wrapload {
        int ((*fn)());
} WRAP_LOAD[] = {
        {Load}
};
#endif	// unix

/*=============================================================================
==   Functions Defined
=============================================================================*/
/*
 *   These are the required driver functions (exported)
 */

int DriverOpen( PVD, PVDOPEN, PUINT16 );
int DriverClose( PVD, PDLLCLOSE, PUINT16 );
int DriverInfo( PVD, PDLLINFO, PUINT16 );
int DriverPoll( PVD, PDLLPOLL, PUINT16 );
int DriverQueryInformation( PVD, PVDQUERYINFORMATION, PUINT16 );
int DriverSetInformation( PVD, PVDSETINFORMATION, PUINT16 );
int DriverGetLastError( PVD, PVDLASTERROR );

static int WFCAPI ICADataArrival( PVOID, USHORT, LPBYTE, USHORT );


/*=============================================================================
==   Data
=============================================================================*/


/*
 *  These are returned when we register our hook
 */

static PWD             pWd              = NULL;
static BOOL            fBufferEmpty     = TRUE; /* True if the data buffer is empty */
static PPING           pPing            = NULL; /* pointer to ping data buffer */
static USHORT          uLen	            = 0;    /* length of ping data to write */
static MEMORY_SECTION  memorySection;

static USHORT gusMaxDataSize;             /* Maximum Data Write Size */
static USHORT gusPingCount;               /* Number of times to ping us */
static USHORT gusVirtualChannelNum;       /* Channel number assigned by WD */


/*******************************************************************************
 *
 *  DriverOpen
 *
 *    Called once to set up things.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdOpen (input/output)
 *       pointer to the structure VDOPEN
 *    puiSize (Output)
 *       size of VDOPEN structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *    CLIENT_ERROR_NO_MEMORY - could not allocate data buffer
 *    On other errors, an error code is returned from lower level functions.
 *
 ******************************************************************************/

int
DriverOpen( PVD pVd, PVDOPEN pVdOpen, PUINT16 puiSize )
{
    WDSETINFORMATION   wdsi;
    VDWRITEHOOK        vdwh;
    WDQUERYINFORMATION wdqi;
    OPENVIRTUALCHANNEL OpenVirtualChannel;
    int                rc;
    UINT16             uiSize;

    TRACE(( TC_VD, TT_API1, "VDPING: DriverOpen entered" ));

    fBufferEmpty = TRUE;

    *puiSize = sizeof(VDOPEN);

    /*
     * Get a virtual channel
     */
    wdqi.WdInformationClass = WdOpenVirtualChannel;
    wdqi.pWdInformation = &OpenVirtualChannel;
    wdqi.WdInformationLength = sizeof(OPENVIRTUALCHANNEL);
    OpenVirtualChannel.pVCName = CTXPING_VIRTUAL_CHANNEL_NAME;

    /* uiSize will be set  when we return back from VdCallWd. */
    uiSize = sizeof( WDQUERYINFORMATION);

    rc = VdCallWd( pVd, WDxQUERYINFORMATION, &wdqi, &uiSize  );
    TRACE(( TC_VD, TT_API1, "VDPING: opening channel %s",
            OpenVirtualChannel.pVCName ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDPING: Could not open %s. rc=%d.",
                OpenVirtualChannel.pVCName, rc));
        return( rc );
    }

    gusVirtualChannelNum = OpenVirtualChannel.Channel;

    /*
     *  Used by the engine
     */
    pVdOpen->ChannelMask = (1L << gusVirtualChannelNum);
    pVd->pPrivate   = NULL; /* pointer to private data, if needed */

    /*
     *  Register write hooks for our virtual channel
     */
    vdwh.Type  = gusVirtualChannelNum;
    vdwh.pVdData = pVd;
    vdwh.pProc = ICADataArrival;
    wdsi.WdInformationClass  = WdVirtualWriteHook;
    wdsi.pWdInformation      = &vdwh;
    wdsi.WdInformationLength = sizeof(VDWRITEHOOK);
    uiSize                   = sizeof( WDSETINFORMATION);

    rc = VdCallWd( pVd, WDxSETINFORMATION, &wdsi, &uiSize  );
    TRACE(( TC_VD, TT_API2, "VDPING: writehook channel number=%u"
            " vdwh.pVdData=%x rc=%d", vdwh.Type, vdwh.pVdData, rc ));

    if ( rc != CLIENT_STATUS_SUCCESS ) {
        TRACE(( TC_VD, TT_ERROR, "VDPING: Could not register write hook. rc %d",
                rc));
        return( rc );
    }

    /*
     *  Grab pointers to functions to use to send data to the host
     */
    pWd                 = vdwh.pWdData;

    TRACE(( TC_VD, TT_API1, "VDPING: Registered"));

    /*
     *  ALL MEMORY MUST BE ALLOCATED DURING INITIALIZATION.
     *  Allocate a single buffer to respond to the ping request.
     *  This example shows use of the MaximumWriteSize returned via
     *  the previous call.
     *  Subtract one because the first byte is used internally by the
     *  WD for the channel number.  We are given a pointer to the
     *  second byte.
     */
    gusMaxDataSize = vdwh.MaximumWriteSize - 1;
    if ( (pPing = (PPING)malloc( gusMaxDataSize )) == NULL ) {
        return( CLIENT_ERROR_NO_MEMORY );
    }

    /*
     * Get tunable parameters from the Ping section of the MODULE.INI file.
     */

    gusPingCount = (USHORT)miGetPrivateProfileInt(INI_PING_VDSECTION,
                                                  INI_PING_PINGCOUNT,
                                                  DEF_PING_PINGCOUNT);

    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  ICADataArrival
 *
 *   A data PDU arrived over our channel.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *
 *    uChan (input)
 *       ICA channel the data is for.
 *
 *    pBuf (input)
 *       Buffer with arriving data packet
 *
 *    Length (input)
 *       Length of the data packet
 *
 * EXIT:
 *       void
 *
 ******************************************************************************/

static int WFCAPI ICADataArrival( PVOID pVd, USHORT uChan, LPBYTE pBuf,
                                  USHORT Length )
{
    int rc;

    WIRE_PTR(PING, pPacket);

    WIRE_READ(PING, pPacket, pBuf);

    TRACE(( TC_VD, TT_API3,
        "VDPING: IcaDataArrival, Len=%d, uLen=%d, pData=[%x][%x][%x][%x]",
        Length, pPacket->uLen, ((LPBYTE)pPacket)[0], ((LPBYTE)pPacket)[1], ((LPBYTE)pPacket)[2], ((LPBYTE)pPacket)[3] ));
    TRACEBUF(( TC_VD, TT_API3, pPacket, Length));

    /*
     *  This protocol is completely synchronous - host should not send
     *  another message with a pending response.
     */
    if ( !fBufferEmpty ) {
        TRACE(( TC_VD, TT_ERROR,
            "VDPING: ICADataArrival - Error: not all data was sent" ));
        return CLIENT_STATUS_SUCCESS;
    }

    /*
     *  Fill in response: echo back everything that server sent us, to a point
     */
    if (pPacket->uLen > gusMaxDataSize) {
        pPing->uLen = gusMaxDataSize;
    }
    else {
        pPing->uLen = pPacket->uLen;
    }

    memcpy( pPing, pPacket, pPing->uLen );
    pPing->ulClientMS = Getmsec();
    fBufferEmpty = FALSE;
    TRACE(( TC_VD, TT_API3,
        "VDPING: ICADataArrival - data queued up, len %d ",
        pPing->uLen ));
    TRACEBUF(( TC_VD, TT_API3, pPing, pPing->uLen ));

    /*
     * save the length, because WIRE_WRITE byte-swaps it
     */
    uLen = pPing->uLen;
    WIRE_WRITE(PING, pPing, pPing->uLen);

    /* Immediately send the return Ping packet */
    memorySection.length = uLen;
    memorySection.pSection = (LPBYTE) pPing;

    rc = QueueVirtualWrite(pWd, gusVirtualChannelNum, &memorySection, 1, FLUSH_IMMEDIATELY);

    if ( rc == CLIENT_STATUS_SUCCESS ) 
        fBufferEmpty = TRUE;

    return CLIENT_STATUS_SUCCESS;
}


/*******************************************************************************
 *
 *  DriverPoll
 *
 *  The Winstation driver calls DriverPoll
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdPoll (input)
 *       pointer to the structure DLLPOLL  (not used)
 *    puiSize (input)
 *       size of DLLPOOL structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - OK so far.  We will be polled again.
 *    CLIENT_STATUS_NO_DATA - No more data to send.
 *    CLIENT_STATUS_ERROR_RETRY - Could not get an output buffer.
 *                            Hopefully, one will be available next time.
 *    Otherwise, a fatal error code is returned from lower level functions.
 *
 *    NOTE:  CLIENT_STATUS_NO_DATA signals the WD that it is OK to stop
 *           polling until another host to client packet comes through.
 ******************************************************************************/

int
DriverPoll( PVD pVd, PDLLPOLL pVdPoll, PUINT16 puiSize )
{
    /* We don't require to send data on this poll any more. Data can be sent
       at any point with the QueueVirtualWrite() call. In this case the
       DriverPoll() is used for error recovery if QueueVirtualWrite() was not
       able to complete the first time. */

    if ( fBufferEmpty == FALSE )
    {
        /* QueueVirtualWrite() failed to write the first time. */
        int rc;

        rc = QueueVirtualWrite(pWd, gusVirtualChannelNum, &memorySection, 1, FLUSH_IMMEDIATELY);

        if ( rc == CLIENT_STATUS_SUCCESS )
        {
            fBufferEmpty = TRUE;
            return CLIENT_STATUS_SUCCESS;
        }
    }

    return CLIENT_STATUS_NO_DATA;
}


/*******************************************************************************
 *
 *  DriverClose
 *
 *  The user interface calls VdClose to close a Vd before unloading it.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to procotol driver data structure
 *    pVdClose (input/output)
 *       pointer to the structure DLLCLOSE
 *    puiSize (input)
 *       size of DLLCLOSE structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/
int
DriverClose( PVD pVd, PDLLCLOSE pVdClose, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDPING: DriverClose entered" ));

    free( pPing );
    pPing = NULL;
    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverInfo
 *
 *    This routine is called to get module information
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdInfo (output)
 *       pointer to the structure DLLINFO
 *    puiSize (output)
 *       size of DLLINFO structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverInfo( PVD pVd, PDLLINFO pVdInfo, PUINT16 puiSize )
{
    USHORT ByteCount;
    PVDPING_C2H pVdData;
    PMODULE_C2H pHeader;
    PVDFLOW pFlow;

    ByteCount = sizeof(VDPING_C2H);

    *puiSize = sizeof(DLLINFO);

    TRACE(( TC_VD, TT_API1, "VDPING: DriverInfo entered" ));

    /*
     *  Check if buffer is big enough
     *  If not, the caller is probably trying to determine the required
     *  buffer size, so return it in ByteCount.
     */
    if ( pVdInfo->ByteCount < ByteCount ) {
        pVdInfo->ByteCount = ByteCount;
        return( CLIENT_ERROR_BUFFER_TOO_SMALL );
    }

    /*
     *  Initialize default data
     */
    pVdInfo->ByteCount = ByteCount;
    pVdData = (PVDPING_C2H) pVdInfo->pBuffer;

    /*
     *  Initialize module header
     */
    pHeader = &pVdData->Header.Header;
    pHeader->ByteCount = ByteCount;
    pHeader->ModuleClass = Module_VirtualDriver;

    pHeader->VersionL = CTXPING_VER_LO;
    pHeader->VersionH = CTXPING_VER_HI;
    strcpy((char*)(pHeader->HostModuleName), "ICA"); /* max 8 characters */

    /*
     *  Initialize virtual driver header
     */
    pVdData->Header.ChannelMask = pVd->ChannelMask;
    pFlow = &pVdData->Header.Flow;
    pFlow->BandwidthQuota = 0;
    pFlow->Flow = VirtualFlow_None;

    /* add our own data */
    pVdData->usMaxDataSize = gusMaxDataSize;
    pVdData->usPingCount = gusPingCount;

    pVdInfo->ByteCount = WIRE_WRITE(VDPING_C2H, pVdData, ByteCount);

    return( CLIENT_STATUS_SUCCESS );
}



/*******************************************************************************
 *
 *  DriverQueryInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdQueryInformation (input/output)
 *       pointer to the structure VDQUERYINFORMATION
 *    puiSize (output)
 *       size of VDQUERYINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverQueryInformation( PVD pVd, PVDQUERYINFORMATION pVdQueryInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDPING: DriverQueryInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdQueryInformation->VdInformationClass = %d",
         pVdQueryInformation->VdInformationClass));

    *puiSize = sizeof( VDQUERYINFORMATION );

    return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverSetInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdSetInformation (input/output)
 *       pointer to the structure VDSETINFORMATION
 *    puiSize (input)
 *       size of VDSETINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int
DriverSetInformation( PVD pVd, PVDSETINFORMATION pVdSetInformation, PUINT16 puiSize )
{
    TRACE(( TC_VD, TT_API1, "VDPING: DriverSetInformation entered" ));
    TRACE(( TC_VD, TT_API2, "pVdSetInformation->VdInformationClass = %d",
         pVdSetInformation->VdInformationClass));

   return( CLIENT_STATUS_SUCCESS );
}


/*******************************************************************************
 *
 *  DriverGetLastError
 *
 *   Queries error data.
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pLastError (output)
 *       pointer to the error structure to return (message is currently always
 *       NULL)
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverGetLastError( PVD pVd, PVDLASTERROR pLastError )
{
    TRACE(( TC_VD, TT_API1, "VDPING: DriverGetLastError entered" ));
    /*
     *  Interpret last error and pass back code and string data
     */

    pLastError->Error = pVd->LastError;
    pLastError->Message[0] = '\0';
    return(CLIENT_STATUS_SUCCESS);
}

